import io
import os
import json
import numpy as np
import pandas as pd
import tensorflow as tf
from keras import backend as K
from keras.models import Sequential
from keras.layers.core import Dense, Activation, Lambda
from keras.layers.recurrent import LSTM

flags = tf.app.flags
FLAGS = flags.FLAGS
flags.DEFINE_integer('time_steps', 24, 'Number of time steps.')
flags.DEFINE_integer('hidden_units', 64, 'Number of LSTM hidden units.')
flags.DEFINE_integer('epoch', 300, 'Number of epoch to runs.')
flags.DEFINE_string('path', 'access.csv', 'Path of input CSV file.')

def load_dataset(path, time_steps):
    csv = pd.read_csv(path).iloc[:,1:]
    dataset = csv.values.astype('float32')
    min, max = normalize_params(dataset[:,0].reshape(-1, 1))
    X, y = [], []
    for i in range(0, len(dataset)-time_steps):
        X.append(dataset[i:i+time_steps])
        y.append(dataset[i+time_steps][0])
    return np.array(X).reshape(-1, time_steps, dataset.shape[1]), np.array(y).reshape(-1, 1), min, max

def normalize_params(dataset):
    min = np.min(np.abs(dataset))
    max = np.max(np.abs(dataset-min))
    return min, max

def tf_normalize(inputs, min, max):
    return tf.truediv(tf.subtract(inputs, min), max)

def normalize(dataset, min, max):
    return (dataset-min)/max

def split_dataset(X, y, test_ratio):
    index = int(len(X) * (1-test_ratio))
    return np.split(X, [index], axis=0), np.split(y, [index], axis=0)

def rmse(y_pred, y_true):
    return np.sqrt(((y_true - y_pred) ** 2).mean())

def run_training():
    X, y, min, max = load_dataset(FLAGS.path, FLAGS.time_steps)
    input_size = X.shape[2]
    if input_size > 1:
        mat_min = np.zeros([1, FLAGS.time_steps, input_size], dtype=np.float32)
        mat_min[:,:,0] = min
        mat_max = np.full((1, FLAGS.time_steps, input_size), 1.0, dtype=np.float32)
        mat_max[:,:,0] = max
    (X_train, X_test), (y_train, y_test) = split_dataset(X, y, 0.1)

    with tf.Session() as sess:
        # train
        K.set_session(sess)
        model = Sequential()
        model.add(Lambda(tf_normalize, arguments={
                    'min': min if input_size == 1 else mat_min,
                    'max': max if input_size == 1 else mat_max
                  }, batch_input_shape=(None, FLAGS.time_steps, input_size)))
        model.add(LSTM(FLAGS.hidden_units, return_sequences=False))
        model.add(Dense(1))
        model.add(Activation("linear"))
        model.compile(loss="mean_squared_error", optimizer="adam")
        model.fit(X_train, normalize(y_train, min, max), epochs=FLAGS.epoch, validation_split=0.05)

        # predict
        pred = model.predict(X_test)
        print("RMSE: {}".format(rmse(pred, normalize(y_test, min, max))))

def main(_):
    run_training()

if __name__ == '__main__':
    tf.app.run()
